#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* Configuration options */

/* Size of each stack frame */
#define STACK_SIZE 256

/* Type to use for stack cells. Should be >= fungespace cells in size */
#define STACK_TYPE signed long

/* Size of each fungespace cell block */
/* e.g. 16 produces 16x16 blocks */
/* Must be power of 2 */
#define CELL_SIZE 16

/* Type to use for fungespace cells */
#define CELL_TYPE signed long

/* Path seperator */
#ifndef PATH_SEPERATOR
#ifdef __riscos__
#define PATH_SEPERATOR '.'
#else
#ifdef __MSDOS__
#define PATH_SEPERATOR '\\'
#else
#define PATH_SEPERATOR '/'
#endif
#endif
#endif

/* Interpreter version (*100) */
#define VERSION 201

/* Interpreter handprint 'BEFC' */
#define HANDPRINT 0x42454643

/* Stack/stackstack system */

typedef struct _bstack {
	STACK_TYPE stack[STACK_SIZE];
	struct _bstack *next;
} bstack;

typedef struct _bstackstack {
	bstack *stack; /* Stack frames */
	signed long pos; /* Location of next free entry in current stack frame */
	struct _bstackstack *next;
} bstackstack;

bstackstack *stack = (bstackstack *) NULL;

STACK_TYPE popcell(signed long n) /* pop from stack n, 0=toss,1=soss, etc. */
{
	bstackstack *ss;
	bstack *s;
	if ((stack == 0) || (n < 0))
		return 0;
	ss = stack;
	while ((n-- > 0) && (ss != NULL))
		ss = ss->next;
	if (ss == NULL)
		return 0;
	if (ss->pos == 0)
	{
		if ((ss->stack) && (ss->stack->next)) /* Any more stack frames? */
		{
			s = ss->stack;
			ss->stack = ss->stack->next;
			ss->pos+=STACK_SIZE;
			free(s);
		}
		else
			return 0; /* No stack */
	}
	ss->pos--;
	return ss->stack->stack[ss->pos];
}

void pushcell(signed long n,STACK_TYPE v) /* push onto stack n */
{
	bstackstack *ss;
	bstack *s;
	if (n < 0)
		return;
	ss = stack;
	while ((n-- > 0) && (ss != NULL))
		ss = ss->next;
	if (ss == NULL)
		return;
	if (ss->stack == NULL)
	{
		ss->stack = (bstack *) malloc(sizeof(bstack));
		ss->stack->next = 0;
	}
	ss->stack->stack[ss->pos++] = v;
	if (ss->pos != STACK_SIZE)
		return;
	ss->pos-=STACK_SIZE;
	s = (bstack *) malloc(sizeof(bstack));
	s->next = ss->stack;
	ss->stack = s;
}

STACK_TYPE peekcell(signed long n,signed long o) /* Peek a value from stack n. offset 0 is top */
{
	bstackstack *ss;
	bstack *s;
	if ((stack == 0) || (n < 0))
		return 0;
	ss = stack;
	while ((n-- > 0) && (ss != NULL))
		ss = ss->next;
	if (ss == NULL)
		return 0;
	s = ss->stack;
	o = ss->pos-(o+1);
	/* Need to make ofs positive */
	while ((o < 0) && (s->next))
	{
		s = s->next;
		o += STACK_SIZE;
	}
	if ((o < 0) && (s->next == NULL)) /* Gone off bottom of stack? */
		return 0;
	return s->stack[o];
}

void pushstack() /* Add new stack to stackstack */
{
	bstackstack *ss;
	ss = (bstackstack *) malloc(sizeof(bstackstack));
	ss->next = stack;
	ss->pos = 0;
	ss->stack = (bstack *) 0;
	stack = ss;
}

void popstack() /* Pop a stack from stackstack */
{
	bstackstack *ss;
	ss = stack;
	if (ss == NULL)
		return;
	while ((ss->stack != NULL) && (ss->stack->next != NULL))
		popcell(0); /* Pop cells until only one frame left. slow but simple ;) */
	stack = ss->next;
	if (ss->stack != NULL)
		free(ss->stack);
	free(ss);
}

signed long stacksize(signed long n) /* return size of stack n */
{
	bstackstack *ss;
	bstack *s;
	signed long cnt;
	if (n < 0)
		return 0;
	ss = stack;
	while ((n-- > 0) && (ss != NULL))
		ss = ss->next;
	if (ss == NULL)
		return 0;
	s = ss->stack;
	cnt = ss->pos;
	if (s == NULL)
		return 0;
	s = s->next;
	while (s)
	{
		cnt += STACK_SIZE;
		s = s->next;
	}
	return cnt;
}

signed long stackcount() /* Return number of stacks on stackstack */
{
	signed long cnt;
	bstackstack *ss;
	cnt = 0;
	ss = stack;
	while (ss)
	{
		ss = ss->next;
		cnt++;
	}
	return cnt;
}

void pushstring(signed long n,char *str) /* Push string to stack n */
{
	signed long p;
	p=0;
	while (str[p])
		p++;
	pushcell(n,0);
	while (p--)
		pushcell(n,str[p]);
}

void popstring(signed long n,char *str) /* Pop string from stack n, assumes str is long enough */
{
	signed long p;
	p=0;
	do {
		str[p] = popcell(n);
	} while (str[p++]);
}

/* Just to make life easier */

void pushvector(signed long n,STACK_TYPE x,STACK_TYPE y)
{
	pushcell(n,x);
	pushcell(n,y);
}

void popvector(signed long n,STACK_TYPE *x,STACK_TYPE *y)
{
	*y = popcell(n);
	*x = popcell(n);
}

/* Memory cell system */

/* Quick explanation:
   This is a simplified version of the system used by !Befunge
   It implements the memory as a binary search tree, with the data items held at the leaves
   Each data item is either another tree, or a bcell (block of cells)
   e.g. the root tree is for searching in the X direction. Its data items will be a set of trees for searching in the Y direction. Their data items will be the cells themselves
   .mask is a bit mask to be applied to the current scalar (called 'check').
         E.g. a mask of 0xFF00 will check the top 8 bits of 'check' (for 16 bit
         integers, of course). If (mask & check) != pos then the required node
         does not exist
   .twiddle is a 'twiddle bit', a power of 2 smaller than the mask (e.g. 0x0080)
            This decides which child node the search will continue to, e.g.
            if ((twiddle & check) == 1) then go to 'one' branch else 'zero' branch
   .pos is the location this node represents. E.g. for a mask of 0xFF00 and a
        position of 0x1200, this node contains the values 0x1200 to 0x12FF
   The end of a tree always has a twiddle value equal to that of CELL_SIZE, since CELL_SIZE cells exist in each bcell and so subdividing further is pointless.
   The mask difference between each node level in the tree need not be a single bit, allowing empty areas to be skipped; the start tree is simply a node of mask 0xFFE0 (for CELL_SIZE 16) and position 0x0000; if access to an address such as 0x0123 is needed then the mask value will be crept back to 0xFE00, and a new node inserted ('zero' child going to the 0xFFE0 node, and 'one' child going to a leaf node with position 0x0120).
   This system means that each tree is no more than sizeof(STACK_TYPE)*8 levels deep (in reality even less since bcells are larger than 1x1), and so requires no balancing. It will be a bit slower than a 'proper' binary search tree because of this, but not by much.
   Deleting empty bcells isn't implemented (but isn't too hard to do)
*/

typedef struct {
	CELL_TYPE cells[CELL_SIZE][CELL_SIZE]; /* [x][y] */
} bcell;

typedef struct {
	STACK_TYPE mask,twiddle,pos;
	void *zero,*one; /* Either another bmem or a bcell */
} bmem;

bmem *cells = (bmem *) NULL;
STACK_TYPE minx,miny,maxx,maxy; /* min/max locations written to */
STACK_TYPE ofsx,ofsy; /* Storage offsets */

STACK_TYPE _readspace(STACK_TYPE x,STACK_TYPE y,bcell *cell)
{
	if (cell == NULL)
		return 32;
	/* Else look it up */
	return cell->cells[(CELL_SIZE-1) & x][(CELL_SIZE-1) & y];
}

bcell *_emptycell()
{
	/* Create empty cell & return */
	STACK_TYPE x,y;
	bcell *b;
	b = malloc(sizeof(bcell));
	for (x=0;x<CELL_SIZE;x++)
		for (y=0;y<CELL_SIZE;y++)
			b->cells[x][y] = 32;
	return b;
}

void _setspace(STACK_TYPE x,STACK_TYPE y,bcell *cell,STACK_TYPE v)
{
	cell->cells[(CELL_SIZE-1) & x][(CELL_SIZE-1) & y] = v;
}

STACK_TYPE readspace(STACK_TYPE x,STACK_TYPE y)
{
	bmem *branch;
	STACK_TYPE check;
	int dim; /* Dimension we are searching - 0=x, 1=y */
	dim = 0;
	check = x;
	if ((x < minx) || (x > maxx) || (y < miny) || (y > maxy))
		return 32;
	branch = cells;
	do {
		if ((check & branch->mask) != branch->pos)
			return 32; /* No bcell for that position */
		/* Else go down a branch */
		if (check & branch->twiddle)
		{
			/* Go down 'one' */
			if (branch->twiddle == CELL_SIZE) /* End of tree? */
			{
				if (dim == 1)
					return _readspace(x,y,(bcell *) branch->one);
				dim = 1;
				check = y;
			}
			branch = (bmem *) branch->one;
		}
		else
		{
			/* Go down 'zero' */
			if (branch->twiddle == CELL_SIZE)
			{
				if (dim == 1)
					return _readspace(x,y,(bcell *) branch->zero);
				dim = 1;
				check = y;
			}
			branch = (bmem *) branch->zero;
		}
	} while (branch);
	return 32;
}

void setspace(STACK_TYPE x,STACK_TYPE y,STACK_TYPE v)
{
	bmem *branch,*newb;
	int dim;
	STACK_TYPE check;
	if (x < minx)
		minx = x;
	else if (x > maxx)
		maxx = x;
	if (y < miny)
		miny = y;
	else if (y > maxy)
		maxy = y;
	dim = 0;
	check = x;
	branch = cells;
	newb = 0;
	do {
		if ((check & branch->mask) != branch->pos)
		{
			/* Insert new branch & empty dimensions */
			/* Need to identify the first bit which differs from the mask */
			newb = malloc(sizeof(bmem));
			newb->mask = branch->mask; /* Make copy of 'branch' so 'branch' can become new branch (to avoid modifying 'branch's parent to point to the new node) */
			newb->twiddle = branch->twiddle;
			newb->pos = branch->pos;
			newb->zero = branch->zero;
			newb->one = branch->one;
			while ((check & branch->mask) != (branch->pos & branch->mask))
			{ /* while masked check value doesn't match masked position of node */
				branch->twiddle *= 2; /* Increase twiddle bit */
				branch->mask ^= branch->twiddle; /* Clear twiddle bit from mask */
			}
			branch->pos = branch->pos & branch->mask; /* Forget lower bits of position */
			/* Now relink */
			if (check & branch->twiddle)
			{
				/* Storing in branch one */
				branch->zero = newb; /* Original branch */
				branch->one = newb = malloc(sizeof(bmem));
			}
			else
			{
				branch->one = newb;
				branch->zero = newb = malloc(sizeof(bmem));
			}
			/* Code at end of function will fill in the new leaves */
		}
		else if (check & branch->twiddle) /* Else go down a branch */
		{
			/* Go down 'one' */
			if (branch->twiddle == CELL_SIZE)
			{
				if (dim == 1)
				{
					if (!(branch->one))
						branch->one = _emptycell();
					_setspace(x,y,(bcell *) branch->one,v);
					return;
				}
				dim = 1;
				check = y;
			}
			if (!(branch->one))
				branch->one = newb = malloc(sizeof(bmem)); /* Create empty dimensions */
			branch = (bmem *) branch->one;
		}
		else
		{
			/* Go down 'zero' */
			if (branch->twiddle == CELL_SIZE)
			{
				if (dim == 1)
				{
					if (!(branch->zero))
						branch->zero = _emptycell();
					_setspace(x,y,(bcell *) branch->zero,v);
					return;
				}
				dim = 1;
				check = y;
			}
			if (!(branch->zero))
				branch->zero = newb = malloc(sizeof(bmem)); /* Create empty dimensions */
			branch = (bmem *) branch->zero;
		}
	} while (!newb);
	/* Fill in newb with the extra dimensions */
	/* i.e. make newb a terminating branch, and make its zero/one child another tree or a cell */
	/* Termiating branches have twiddle = CELL_SIZE, mask = ~((CELL_SIZE*2)-1), pos = check & mask */
	do {
		newb->mask = ~((CELL_SIZE*2)-1);
		newb->twiddle = CELL_SIZE;
		newb->pos = check & newb->mask;
		if (check & newb->twiddle)
		{
			/* We are writing to block 1 */
			newb->zero = 0;
			if (dim == 1)
			{
				/* End */
				newb->one = _emptycell();
				_setspace(x,y,(bcell *) newb->one,v);
				return;
			}
			/* Else more dimensions needed */
			dim = 1;
			check = y;
			newb->one = malloc(sizeof(bmem));
			newb = newb->one;
		}
		else
		{
			newb->one = 0;
			if (dim == 1)
			{
				/* End */
				newb->zero = _emptycell();
				_setspace(x,y,(bcell *) newb->zero,v);
				return;
			}
			/* Else more dimensions needed */
			dim = 1;
			check = y;
			newb->zero = malloc(sizeof(bmem));
			newb = newb->zero;
		}
	} while (1);
}

/* File access */

int load_file(char *name,STACK_TYPE bx,STACK_TYPE by,int flags,STACK_TYPE *sx,STACK_TYPE *sy)
{
	FILE *f;
	STACK_TYPE px,py,c,oc;
	if (flags * 1)
		f = fopen(name,"rb");
	else
		f = fopen(name,"r");
	if (!f)
		return EOF; /* error! */
	px = py = *sx = *sy = oc = 0;
	c = fgetc(f);
	while (c != EOF)
	{
		if (flags & 1)
		{
			if (c != 32) /* FBBI says skip spaces in binary mode */
				setspace(bx+px,by+py,c);
			px++;
			if (*sx < px)
				*sx = px;
		}
		else if ((c == 10) || (c == 13))
		{
			if (((c == 10) && (oc != 13)) || ((c == 13) && (oc != 10))) /* Accept 10, 13, 10-13 and 13-10 */
			{
				py++;
				px = 0;
			}
		}
		else
		{
			if (c != 32)
				setspace(bx+px,by+py,c);
			px++;
			if (*sx < px)
				*sx = px;
		}
		oc = c;
		c = fgetc(f);
	}
	fclose(f);
	*sy = py+1; /* newlines+1 */
	return 0; /* OK */
}

int save_file(char *name,STACK_TYPE bx,STACK_TYPE by,STACK_TYPE sx,STACK_TYPE sy,int flags)
{
	FILE *f;
	STACK_TYPE px,py,ox,oy,c;
	f = fopen(name,"w");
	if (!f)
		return EOF; /* error */
	oy = px = py = 0;
	if (sy > 0)
	do
	{
		ox = 0;
		if (sx > 0)
		do
		{
			c = readspace(bx+px,by+py);
			if ((c != 32) || ((flags & 1) == 0))
			{
				while (oy < py)
				{
					if (fprintf(f,"\n") == EOF)
					{
						fclose(f);
						return EOF;
					}
					oy++;
				}
				while (ox < px)
				{
					if (fputc(32,f) == EOF)
					{
						fclose(f);
						return EOF;
					}
					ox++;
				}
				if (fputc(c,f) == EOF)
				{
					fclose(f);
					return EOF;
				}
				ox++;
			}
			px++;
		} while (px < sx);
		px=0;
		py++;
		if ((py < sy) && ((flags & 1) == 0)) /* more Y left, and not compressed */
		{
			if (fprintf(f,"\n") == EOF)
			{
				fclose(f);
				return EOF;
			}
			oy++;
		}
	} while (py < sy);
	fclose(f);
	return 0; /* OK */
}

/* y function */

void yfunc(int argc,char **argv,STACK_TYPE x,STACK_TYPE y,STACK_TYPE xd,STACK_TYPE yd)
{
	STACK_TYPE yval;
	signed long ss,tmp,tmp2;
	struct tm *thetime;
	time_t atime;
	yval = popcell(0); /* Get y-value */
	ss = stacksize(0); /* Size of current stack */
	/* 20: Environment variables */
	pushcell(0,0);
	/* 19: Command line args */
	pushcell(0,0); /* Double null */
	pushcell(0,0);
	for (tmp=argc-1;tmp>=0;tmp--)
	{
		if (argv[tmp] == 0)
			pushcell(0,0); /* Just in case */
		else
			pushstring(0,argv[tmp]);
	}
	/* 18: Stack sizes */
	tmp = tmp2 = stackcount();
	while (tmp-- > 1)
		pushcell(0,stacksize(tmp));
	pushcell(0,ss); /* Size of stack 0 */
	/* 17: Number of stacks */
	pushcell(0,tmp2);
	/* 16: Current time */
	atime = time(NULL);
	thetime = localtime(&atime);
	pushcell(0,(thetime->tm_hour*65536)+(thetime->tm_min*256)+thetime->tm_sec);
	/* 15: Current date */
	pushcell(0,(thetime->tm_year*65536)+(thetime->tm_mon*256)+thetime->tm_mday);
	/* 14: Greatest non-cell pos, relative to least */
	pushvector(0,maxx-minx,maxy-miny);
	/* 13: Least non-cell pos */
	pushvector(0,minx,miny);
	/* 12: Storage offset */
	pushvector(0,ofsx,ofsy);
	/* 11: Delta */
	pushvector(0,xd,yd);
	/* 10: Pos */
	pushvector(0,x,y);
	/* 9: IP team */
	pushcell(0,0);
	/* 8: IP id */
	pushcell(0,0);
	/* 7: No. of dims */
	pushcell(0,2);
	/* 6: Path seperator */
	pushcell(0,PATH_SEPERATOR);
	/* 5: Operating paradigm */
	pushcell(0,1); /* Like C system() */
	/* 4: Interpreter version */
	pushcell(0,VERSION);
	/* 3: Interpreter handprint */
	pushcell(0,HANDPRINT);
	/* 2: Bytes per cell */
	pushcell(0,sizeof(CELL_TYPE));
	/* 1: Flags */
	pushcell(0,2+4+8);
	/* Now handle y-value */
	if (yval > 0)
	{
		tmp = peekcell(0,yval-1);
		tmp2 = stacksize(0);
		while (tmp2-- > ss)
			popcell(0);
		pushcell(0,tmp);
	}
}


/* Initialisation */

void initialise()
{
	signed long x,y;
	/* Set up stacks */
	pushstack();
	/* Set up memory */
	minx = miny = maxx = maxy = ofsx = ofsy = 0;
	/* Initial leaf node */
	cells = malloc(sizeof(bmem));
	cells->mask = ~((CELL_SIZE*2)-1);
	cells->twiddle = CELL_SIZE;
	cells->pos = 0;
	cells->zero = cells->one = 0;
	/* Copy initial fungespace from char **initfunge */
	y=0;
	while (initfunge[y][0])
	{
		x=0;
		while (initfunge[y][x])
		{
			setspace(x,y,initfunge[y][x]);
			x++;
		}
		y++;
	}
}
